/*******************************************************************************
        
        copyright:      Copyright (C) Mike Reinhold 2010. All rights reserved
                        
        license:        BSD style: $(LICENSE)
                
        author(s):      Mike Reinhold
                        
*******************************************************************************/
module tango.util.cipher.modes.CFB;

import tango.util.cipher.Cipher;

/** This class implements the cipher feed back (CFB) block mode */
class CFB: BlockCipherMode {
    private {
        BlockCipher wrappedCipher;
        ubyte[] iv, counter;
    }

    /** Null constructor */
    this() {
    }

    /***************************************************************************
    
            constructor
            
            Params:
            cipher = block cipher to wrap

    ***************************************************************************/
    this(BlockCipher cipher) {
        wrappedCipher = cipher;
    }

    
    /***************************************************************************
    
            cipher
            
            Returns:
            the underlying cipher we are wrapping.

    ***************************************************************************/
    BlockCipher cipher() {
        return wrappedCipher;
    }

    /***************************************************************************
    
            cipher
            
            Params:
            cipher_ = block cipher to wrap
            
            Remarks:
            Mutator method which sets the wrapped block cipher.

    ***************************************************************************/
    void cipher(BlockCipher cipher_) {
        wrappedCipher = cipher_;
    }

    /***************************************************************************
    
            name
            
            Returns:
            the name of the wrapped cipher and its mode.
            
            Remarks:
            If this were wrapping Serpent the result would be "Serpent/CFB"

    ***************************************************************************/
    char[] name() {
        return wrappedCipher.name ~ "/CFB";
    }
    
    /***************************************************************************
    
            requiresPadding
            
            Returns:
            does this mode require padding?

    ***************************************************************************/
    bool requiresPadding(){
        return false;
    }

    /***************************************************************************
    
            init
            
            Params:
            encrypt = the direction of the cipher
            key =     symmetric key to be used
            
            Remarks:
            A new IV is generated by this method

    ***************************************************************************/
    void init(bool encrypt, SymmetricKey key) {
        init(encrypt, key, new IVParameters());
    }

    /***************************************************************************
    
            init
            
            Params:
            encrypt =  the direction of the cipher
            key =      symmetric key to be used
            ivParams = the initialization vector to be used
            
            Remarks:
            If the parameters are incorrect an exception will be thrown.

    ***************************************************************************/
    void init(bool encrypt, SymmetricKey key, IVParameters ivParams) {
        if(!ivParams)
            invalid(name() ~ ": Block mode requires IV (use ParametersWithIV)");
        if(ivParams.iv.length < blockSize)
            invalid(name() ~ ": IV must be at least as long as cipher block size");

        //initialize the wrapped cipher
        wrappedCipher.init(true, key);

        //prepare the initialization vector and state counter
        iv = ivParams.iv[0 .. blockSize];
        counter = new ubyte[blockSize];
        counter[] = iv;

        _encrypt = encrypt;
        _initialized = true;
    }

    /***************************************************************************
    
            update
            
            Params:
            input_ = the block to be processed
            output_ = the output buffer
                        
            Returns:
            the number of bytes processed
            
            Remarks:
            Update the input block based on the current cipher direction and
                place it in the output block.

    ***************************************************************************/
    uint update(void[] input_, void[] output_) {
        if(!_initialized)
            invalid(name() ~ ": Block mode not initialized");
        ubyte[] input = cast(ubyte[]) input_, 
                output = cast(ubyte[]) output_;

        uint len = (counter.length > input.length) ? input.length : counter.length;

        if(len > output.length)
            invalid(name() ~ ": Output buffer too short");

        //Encrypt the counter
        wrappedCipher.update(counter, counter);

        //XOR output with plaintext to create ciphertext
        for(int i = 0; i < len; i++)
            counter[i] ^= input[i];

        //store the output
        output[0 .. len] = counter[0 .. len];
        
        //set the counter
        if(!_encrypt){
            counter[0..len] = input[0..len];
        }
        
        //return the number of bytes processed
        return len;
    }

    /***************************************************************************
    
            blockSize
                                            
            Returns:
            the size of a block as specified by the wrapped cipher

    ***************************************************************************/
    uint blockSize() {
        return wrappedCipher.blockSize;
    }

    /***************************************************************************
    
            ivSize
                                            
            Returns:
            the length of the initialization vector in bytes

    ***************************************************************************/
    uint ivSize() {
        return wrappedCipher.ivSize;
    }

    /***************************************************************************
    
            validKeySizes
                                            
            Returns:
            an array containing the byte length of every possible valid key

    ***************************************************************************/
    uint[] validKeySizes() {
        return wrappedCipher.validKeySizes;
    }

    /***************************************************************************
    
            reset
                                            
            Returns:
            resets the BlockCipherMode and the wrapped cipher

    ***************************************************************************/
    void reset() {
        counter[] = iv;
        wrappedCipher.reset();
    }
}
