/*******************************************************************************
        
        copyright:      Copyright (C) Mike Reinhold 2010. All rights reserved
                        
        license:        BSD style: $(LICENSE)
                
        author(s):      Mike Reinhold
                        
*******************************************************************************/
module tango.util.cipher.modes.ECB;

import tango.util.cipher.Cipher;

/** This class implements the electronic code book(ECB) block mode. */
class ECB: BlockCipherMode {
    private {
        BlockCipher wrappedCipher;
    }

    /** Null constructor */
    this() {
    }

    /***************************************************************************
    
            constructor
            
            Params:
            cipher = block cipher to wrap

    ***************************************************************************/
    this(BlockCipher cipher) {
        wrappedCipher = cipher;
    }

    /***************************************************************************
    
            cipher
            
            Returns:
            the underlying cipher we are wrapping.

    ***************************************************************************/
    BlockCipher cipher() {
        return wrappedCipher;
    }

    /***************************************************************************
    
            cipher
            
            Params:
            cipher_ = block cipher to wrap
            
            Remarks:
            Mutator method which sets the wrapped block cipher.

    ***************************************************************************/
    void cipher(BlockCipher cipher_) {
        wrappedCipher = cipher_;
    }

    /***************************************************************************
    
            name
            
            Returns:
            the name of the wrapped cipher and its mode.
            
            Remarks:
            If this were wrapping Serpent the result would be "Serpent/ECB"

    ***************************************************************************/
    char[] name() {
        return wrappedCipher.name ~ "/ECB";
    }
    
    /***************************************************************************
    
            requiresPadding
            
            Returns:
            does this mode require padding?

    ***************************************************************************/
    bool requiresPadding(){
        return true;
    }

    /***************************************************************************
    
            init
            
            Params:
            encrypt = the direction of the cipher
            key =     symmetric key to be used
            
            Remarks:
            A new IV is generated by this method

    ***************************************************************************/
    void init(bool encrypt, SymmetricKey key) {
        init(encrypt, key, new IVParameters());
    }

    /***************************************************************************
    
            init
            
            Params:
            encrypt =  the direction of the cipher
            key =      symmetric key to be used
            ivParams = the initialization vector to be used
            
            Remarks:
            If the parameters are incorrect an exception will be thrown.

    ***************************************************************************/
    void init(bool encrypt, SymmetricKey key, IVParameters ivParams) {
        if(!ivParams)
            invalid(name() ~ ": Block mode requires IV (use ParametersWithIV)");
        if(ivParams.iv.length < blockSize)
            invalid(name() ~ ": IV must be at least as long as cipher block size");

        _encrypt = encrypt;

        //initialize the wrapped cipher
        wrappedCipher.init(_encrypt, key);

        _initialized = true;
    }

    /***************************************************************************
    
            update
            
            Params:
            input_ = the block to be processed
            output_ = the output buffer
                        
            Returns:
            the number of bytes processed
            
            Remarks:
            Update the input block based on the current cipher direction and
                place it in the output block.
    
    ***************************************************************************/
    uint update(void[] input_, void[] output_) {
        if(!_initialized)
            invalid(name() ~ ": Block mode not initialized");

        ubyte[] input = cast(ubyte[]) input_, output = cast(ubyte[]) output_;

        if(input.length < blockSize)
            invalid(name() ~ ": Input buffer too short");

        if(output.length < blockSize)
            invalid(name() ~ ": Output buffer too short");

        if(_encrypt) {
            //encrypt the block
            wrappedCipher.update(input, input);

            //store to output
            output[0 .. blockSize] = input[0 .. blockSize];
        } else {
            ubyte[] temp = new ubyte[input.length];

            //decrypt the block
            wrappedCipher.update(input, temp);

            //store output for block
            output[0 .. blockSize] = temp[0 .. blockSize];
        }

        //return the number of blocks processed
        return blockSize;
    }

    /***************************************************************************
    
            blockSize
                                            
            Returns:
            the size of a block as specified by the wrapped cipher

    ***************************************************************************/
    uint blockSize() {
        return wrappedCipher.blockSize;
    }

    /***************************************************************************
    
            ivSize
                                            
            Returns:
            the length of the initialization vector in bytes

    ***************************************************************************/
    uint ivSize() {
        return wrappedCipher.ivSize;
    }

    /***************************************************************************
    
            validKeySizes
                                            
            Returns:
            an array containing the byte length of every possible valid key
             
    ***************************************************************************/
    uint[] validKeySizes() {
        return wrappedCipher.validKeySizes;
    }

    /***************************************************************************
    
            reset
                                            
            Returns:
            resets the BlockCipherMode and the wrapped cipher

    ***************************************************************************/
    void reset() {
        wrappedCipher.reset();
    }
}