/*******************************************************************************
        
        copyright:      Copyright (C) Mike Reinhold 2010. All rights reserved
                        
        license:        BSD style: $(LICENSE)
                
        author(s):      Mike Reinhold (Based on CBC implementation by Thomas Dixon)
                        
*******************************************************************************/
module tango.util.cipher.modes.PCBC;

import tango.util.cipher.Cipher;

/** This class implements the propagated cipher block chaining (PCBC) block mode. */
class PCBC: BlockCipherMode {
    private {
        BlockCipher wrappedCipher;
        ubyte[] iv, previousBlock, currentBlock;
    }

    /** Null constructor */
    this() {
    }

    /***************************************************************************
    
            constructor
            
            Params:
            cipher = block cipher to wrap

    ***************************************************************************/
    this(BlockCipher cipher) {
        wrappedCipher = cipher;
    }

    /***************************************************************************
    
            cipher
            
            Returns:
            the underlying cipher we are wrapping.
     
    ***************************************************************************/
    BlockCipher cipher() {
        return wrappedCipher;
    }

    /***************************************************************************
    
            cipher
            
            Params:
            cipher_ = block cipher to wrap
            
            Remarks:
            Mutator method which sets the wrapped block cipher.

    ***************************************************************************/
    void cipher(BlockCipher cipher_) {
        wrappedCipher = cipher_;
    }

    /***************************************************************************
    
            name
            
            Returns:
            the name of the wrapped cipher and its mode.
            
            Remarks:
            If this were wrapping Serpent the result would be "Serpent/PCBC"

    ***************************************************************************/
    char[] name() {
        return wrappedCipher.name ~ "/PCBC";
    }
    
    /***************************************************************************
    
            requiresPadding
            
            Returns:
            does this mode require padding?

    ***************************************************************************/
    bool requiresPadding(){
        return true;
    }

    /***************************************************************************
    
            init
            
            Params:
            encrypt = the direction of the cipher
            key =     symmetric key to be used
            
            Remarks:
            A new IV is generated by this method

    ***************************************************************************/
    void init(bool encrypt, SymmetricKey key) {
        init(encrypt, key, new IVParameters());
    }

    /***************************************************************************
    
            init
            
            Params:
            encrypt =  the direction of the cipher
            key =      symmetric key to be used
            ivParams = the initialization vector to be used
            
            Remarks:
            If the parameters are incorrect an exception will be thrown.

    ***************************************************************************/
    void init(bool encrypt, SymmetricKey key, IVParameters ivParams) {
        if(!ivParams)
            invalid(name() ~ ": Block mode requires IV (use ParametersWithIV)");
        if(ivParams.iv.length < blockSize)
            invalid(name() ~ ": IV must be at least as long as cipher block size");

        _encrypt = encrypt;

        //initialize the wrapped cipher
        wrappedCipher.init(_encrypt, key);

        //setup the state
        iv = ivParams.iv[0 .. blockSize];
        currentBlock = new ubyte[blockSize];
        previousBlock = new ubyte[blockSize];
        previousBlock[] = iv;

        _initialized = true;
    }

    /***************************************************************************
    
            update
            
            Params:
            input_ = the block to be processed
            output_ = the output buffer
                        
            Returns:
            the number of bytes processed
            
            Remarks:
            Update the input block based on the current cipher direction and
                place it in the output block.

    ***************************************************************************/
    uint update(void[] input_, void[] output_) {
        if(!_initialized)
            invalid(name() ~ ": Block mode not initialized");

        ubyte[] input = cast(ubyte[]) input_, 
                output = cast(ubyte[]) output_;

        if(input.length < blockSize)
            invalid(name() ~ ": Input buffer too short");

        if(output.length < blockSize)
            invalid(name() ~ ": Output buffer too short");

        if(_encrypt) {
            //XOR the input with the previous block
            for(uint pos = 0; pos < blockSize; pos++)
                previousBlock[pos] ^= input[pos];

            //encrypt the block
            wrappedCipher.update(previousBlock, previousBlock);

            //store the ciphertext
            for(uint pos = 0; pos < blockSize; pos++)
                output[pos] = previousBlock[pos];
            
            //propagate the cipher
            for(uint pos = 0; pos < blockSize; pos++)
                previousBlock[pos] ^= input[pos];
        } else {
            ubyte[] temp = input[0 .. blockSize];

            //decrypt the block
            wrappedCipher.update(temp, currentBlock);
            
            //get the plaintext
            for(uint pos = 0; pos < blockSize; pos++)
                output[pos] = (currentBlock[pos] ^ previousBlock[pos]);

            // Store the propagating previous block
            for(uint pos = 0; pos < blockSize; pos++)
                previousBlock[pos] = temp[pos] ^ output[pos];
        }

        //return the number of processed bytes
        return blockSize;
    }

    /***************************************************************************
    
            blockSize
                                            
            Returns:
            the size of a block as specified by the wrapped cipher

    ***************************************************************************/
    uint blockSize() {
        return wrappedCipher.blockSize;
    }

    /***************************************************************************
    
            ivSize
                                            
            Returns:
            the length of the initialization vector in bytes

    ***************************************************************************/
    uint ivSize() {
        return wrappedCipher.ivSize;
    }

    /***************************************************************************
    
            validKeySizes
                                            
            Returns:
            an array containing the byte length of every possible valid key

    ***************************************************************************/
    uint[] validKeySizes() {
        return wrappedCipher.validKeySizes;
    }

    /***************************************************************************
    
            reset
                                            
            Returns:
            resets the BlockCipherMode and the wrapped cipher

    ***************************************************************************/
    void reset() {
        previousBlock[] = iv;
        wrappedCipher.reset();
    }   
}
